/*
	wmkfxapi.c[pp]
*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include "wmkxfapi.h"


int		wmkxfCountFonts(LPFONTXDBCSHDR fxhdr)
{
	int		result;
	
	result = 256;		// number of SBCS fonts 
	if (fxhdr->flag & 1) {
		WORD	st, ed;
		BYTE	cnt;
		
		result = 0;
		for(cnt = 0; cnt < fxhdr->cnt_rgn; cnt++) {
			st = fxhdr->rgn[cnt].rgn_begin;
			ed = fxhdr->rgn[cnt].rgn_end;
			if (st > ed) { result = 0; break; }
			result += (ed - st) + 1;
		}
	}
	
	return result;
}


int		wmkxfQueryTotalSize(LPDWORD total_size, LPFONTXHDR fontxhdr)
{
	LPFONTXDBCSHDR	fxhdr;
	int				fonts, cbfont;

	fxhdr = (LPFONTXDBCSHDR)fontxhdr;
	if (!fxhdr || memcmp(fxhdr->id, FONTX2ID, 6) != 0) return WMKXF_ERR_HDR;
	fonts = wmkxfCountFonts(fxhdr);
	if (!fonts) return WMKXF_ERR_RGN;
	cbfont = (((int)(fxhdr->width) + 7) >> 3) * (int)(fxhdr->height);
	if (cbfont <= 0) return WMKXF_ERR_HDR;
	
	if (total_size != NULL) {
		*total_size = sizeof(FONTXHDR);
		if (fxhdr->flag & 1) {
			*total_size += sizeof(BYTE) + sizeof(FONTXRGN) * fxhdr->cnt_rgn;
		}
		*total_size += fonts * cbfont;
	}
	
	return WMKXF_NOERR;
}


int		wmkxfWriteToMem(LPVOID mem, int length, LPDWORD dw_val)
{
	LPBYTE *	p;
	
	p = (LPBYTE *)dw_val;
	MoveMemory(*p, mem, length);
	*p += length;
	
	return WMKXF_NOERR;
}


int		wmkxfWriteToFile(LPVOID mem, int length, LPDWORD dw_val)
{
	int		result;
	HANDLE	hf;
	DWORD	cb;
	
	hf = *(HANDLE *)dw_val;
	
	result = WriteFile(hf, mem, length, &cb, NULL);
	if (result) result = ((DWORD)length == cb);
	
	return result ? WMKXF_NOERR : WMKXF_ERR_FILE;
}


int		wmkxfOutputFonts(LPFONTXDBCSHDR fx, HFONT hfont, DWORD align_flag, WMKXFWRITEPROC lpfnwrite, LPDWORD dw_val, WMKXFCALLBACKPROC lpfncallback, LPARAM ldata)
{
	int			result;
	WMKXFDATA	wmkdata;
	HDC			hdc;
	HBITMAP		hbmp, hbmp_prev;
	HFONT		hfont_prev;
	COLORREF	fg_prev, bg_prev;
	WORD		font_cnt;
	LPBYTE		dib, tmpfbuf;
	LPBITMAPINFOHEADER	bi;
	int			dib_line;
	int			font_line;
	int			hdr_size;
	RECT		rc;
	int			x_org = 0, y_org = 0;
	
	
	font_cnt = wmkxfCountFonts(fx);
	if (font_cnt <= 0) return WMKXF_ERR_RGN;
	
	hdr_size = sizeof(FONTXHDR);
	if (fx->flag & 1) {
		hdr_size += sizeof(BYTE) + sizeof(FONTXRGN) * fx->cnt_rgn;
	}
	if ((result = lpfnwrite(fx, hdr_size, dw_val)) != WMKXF_NOERR)
		return result;
	
	font_line = ((int)(fx->width) + 7) >> 3;
	tmpfbuf = (LPBYTE)malloc(font_line * fx->height);
	if (!tmpfbuf) return WMKXF_ERR_MEM;
	
	bi = (LPBITMAPINFOHEADER)malloc(sizeof(BITMAPINFOHEADER) + 8);
	if (!bi) return WMKXF_ERR_MEM;
	bi->biSize = sizeof(BITMAPINFOHEADER);
	bi->biWidth = fx->width;
	bi->biHeight = fx->height;
	bi->biPlanes = 1;
	bi->biBitCount = 1;
	bi->biCompression = BI_RGB;
	((LPBITMAPINFO)bi)->bmiColors[0].rgbBlue = 0;
	((LPBITMAPINFO)bi)->bmiColors[0].rgbRed = 0;
	((LPBITMAPINFO)bi)->bmiColors[0].rgbGreen = 0;
	((LPBITMAPINFO)bi)->bmiColors[1].rgbBlue = 255;
	((LPBITMAPINFO)bi)->bmiColors[1].rgbRed = 255;
	((LPBITMAPINFO)bi)->bmiColors[1].rgbGreen = 255;
	
	dib_line = ((int)(bi->biWidth + 31) & (~31)) >> 3;
	rc.top = rc.left = 0;
	rc.bottom = bi->biHeight;
	rc.right = bi->biWidth;
	
	hdc = CreateCompatibleDC(NULL);
	hbmp = CreateDIBSection(hdc,(LPBITMAPINFO)bi,DIB_RGB_COLORS, (LPVOID *)&dib,NULL,0);
	if (hbmp) {
		CHAR	c[2];
		WORD	st, ed;
		WORD	i, j, k, progress;
		WORD	rgncnt;
		LPBYTE	tmpfbuff;
		
		hbmp_prev = (HBITMAP)SelectObject(hdc, hbmp);
		hfont_prev = (HFONT)SelectObject(hdc, hfont);
		fg_prev = SetTextColor(hdc, RGB(255,255,255));
		bg_prev = SetBkColor(hdc, RGB(0, 0, 0));
		if (align_flag & (WMKXF_ALIGN_CENTER)) {
			UINT	align_val = 0;
			
			if (align_flag & WMKXF_ALIGN_CENTER) {
				align_val |= TA_CENTER;
				x_org = bi->biWidth / 2;
			}
			SetTextAlign(hdc, align_val);
		}
		
		progress = 0;
		wmkdata.user_data = ldata;
		wmkdata.total_count = font_cnt;
		
		rgncnt = (fx->flag & 1) ? fx->cnt_rgn : 1;
		for(i=0; i<rgncnt; i++) {
			if (fx->flag & 1) {
				st = fx->rgn[i].rgn_begin;
				ed = fx->rgn[i].rgn_end;
			} else {
				st = 0;
				ed = 255;
			}
			for(j=st; j<=ed; j++) {
				ZeroMemory(dib, bi->biHeight * dib_line);
				if (j < 256)	{ c[0] = (BYTE)j; c[1] = '\0'; }
				else			{ c[0] = HIBYTE(j); c[1] = LOBYTE(j); }
				ExtTextOut(hdc, x_org, y_org, ETO_OPAQUE,
					&rc,
					c,
					(j < 256) ? 1 : 2,
					NULL);
				for(k=0; k < fx->height; k++) {
					MoveMemory(tmpfbuf + font_line * k, 
								dib + (fx->height - k - 1) * dib_line,
								font_line);
				}
				result = lpfnwrite(tmpfbuf, font_line * fx->height, dw_val);
				if (result != WMKXF_NOERR) break;
				progress++;
				if (lpfncallback) {
					wmkdata.progress_code = j;
					wmkdata.progress_count = progress;
					result = lpfncallback(&wmkdata);
					if (result != WMKXF_NOERR) break;
				}
			}
			if (result != WMKXF_NOERR) break;
		}
		
		SetBkColor(hdc, bg_prev);
		SetTextColor(hdc, fg_prev);
		SelectObject(hdc, hfont_prev);
		SelectObject(hdc, hbmp_prev);
		DeleteObject(hbmp);
	}
	DeleteDC(hdc);
	
	if (tmpfbuf) free(tmpfbuf);
	if (bi) free(bi);
	return result;
}


int		wmkxfFontToMem(LPFONTXDBCSHDR fx, HFONT hfont, LPVOID dest, DWORD align_flag, WMKXFCALLBACKPROC lpfncallback, LPARAM ldata)
{
	DWORD	dw;
	
	dw = (DWORD)dest;
	return wmkxfOutputFonts(fx, hfont, align_flag, wmkxfWriteToMem, &dw, lpfncallback, ldata);
}

int		wmkxfFontToFile(LPFONTXDBCSHDR fx, HFONT hfont, HANDLE hf, DWORD align_flag, WMKXFCALLBACKPROC lpfncallback, LPARAM ldata)
{
	DWORD	dw;
	
	dw = (DWORD)hf;
	return wmkxfOutputFonts(fx, hfont, align_flag, wmkxfWriteToFile, &dw, lpfncallback, ldata);
}
