/*
	wmkxfcmd.c[pp]
*/


#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wmkxfapi.h"


static FONTXRGN		rgn_dbcs[] = {
	{ 0x8140, 0x81fc },
	{ 0x8240, 0x82fc },
	{ 0x8340, 0x83fc },
	{ 0x8440, 0x84fc },
	{ 0x8740, 0x87fc },
	{ 0x889f, 0x88fc },  // JIS level 1 
	{ 0x8940, 0x89fc },
	{ 0x8a40, 0x8afc },
	{ 0x8b40, 0x8bfc },
	{ 0x8c40, 0x8cfc },
	{ 0x8d40, 0x8dfc },
	{ 0x8e40, 0x8efc },
	{ 0x8f40, 0x8ffc },
	{ 0x9040, 0x90fc },
	{ 0x9140, 0x91fc },
	{ 0x9240, 0x92fc },
	{ 0x9340, 0x93fc },
	{ 0x9440, 0x94fc },
	{ 0x9540, 0x95fc },
	{ 0x9640, 0x96fc },
	{ 0x9740, 0x97fc },
	{ 0x9840, 0x98fc }, // ...0x9872 level 1, 0x989f... level 2
	{ 0x9940, 0x99fc },
	{ 0x9a40, 0x9afc },
	{ 0x9b40, 0x9bfc },
	{ 0x9c40, 0x9cfc },
	{ 0x9d40, 0x9dfc },
	{ 0x9e40, 0x9efc },
	{ 0x9f40, 0x9ffc },
	{ 0xe040, 0xe0fc },
	{ 0xe140, 0xe1fc },
	{ 0xe240, 0xe2fc },
	{ 0xe340, 0xe3fc },
	{ 0xe440, 0xe4fc },
	{ 0xe540, 0xe5fc },
	{ 0xe640, 0xe6fc },
	{ 0xe740, 0xe7fc },
	{ 0xe840, 0xe8fc },
	{ 0xe940, 0xe9fc },
	{ 0xea40, 0xeaa4 }, // ...0xeaa4 level 2
	
	{ 0xed40, 0xedfc }, // NEC optional
	{ 0xee40, 0xeefc },
	
	{ 0xfa40, 0xfafc }, // IBM expantion
	{ 0xfb40, 0xfbfc },
	{ 0xfc40, 0xfc4b }
};



void		usage(void)
{
	const char		sz[] =
		"wmkxfcmd v0.00\n"
		"wmkxfcmd	Windows ̃tHg FONTX `̃tHg쐬\n"
		"\n"
		"[p@] wmkfxcmd \"FontFace\" [SBCS|DBCS] width height fontname output-name\n"
		"\n"
		"  \"FontFace\"\tWindows ̃tHg\n"
		"  [SBCS|DBCS]\t^Cvip^Spj\n"
		"  width    \t\n"
		"  height   \t̍\n"
		"  fontname \tFONTX tHgip 8 ȓj\n"
		"  output-name\to̓t@C\n"
		"\n"
		"[p]\n"
		"\twmkxfcmd \"lr \" SBCS 8 16 WMKXF wmkhn16f.fnt\n"
		"\tlr  8x16 dot ptHg쐬\n"
		"\n"
		;
		
		puts(sz);
}


LPFONTXDBCSHDR
makeFontxHdr(int w, int h, LPSTR fontxname, BOOL isdbcs)
{
	LPFONTXDBCSHDR	fx;
	int		hdr_size;
	
	hdr_size = sizeof(FONTXHDR);
	if (isdbcs) {
		hdr_size += sizeof(BYTE) + sizeof(rgn_dbcs);
	}
	
	fx = (LPFONTXDBCSHDR)malloc(hdr_size);
	if (fx) {
		size_t		l;
		MoveMemory(fx->id, FONTX2ID, sizeof(fx->id));
		memset(fx->name, ' ', sizeof(fx->name));
		l = lstrlen(fontxname);
		MoveMemory(fx->name, fontxname, min(l, sizeof(fx->name)));
		fx->width = w;
		fx->height = h;
		fx->flag = (isdbcs != 0);
		if (isdbcs) {
			fx->cnt_rgn = sizeof(rgn_dbcs) / sizeof(rgn_dbcs[0]);
			MoveMemory(fx->rgn, rgn_dbcs, sizeof(rgn_dbcs));
		}
	}
	
	return fx;
}


void
disposeFontxHdr(LPVOID *fx)
{
	free(fx);
}


HFONT
makeFont(LPSTR face, int w, int h, BOOL isbold, BOOL isdbcs)
{
	LOGFONT		lf;
	
	ZeroMemory(&lf, sizeof(lf));
	lf.lfHeight = h;
	lf.lfWidth = w >> (isdbcs ? 1 : 0);
	lf.lfWeight = isbold ? 700 : 400;
	lf.lfCharSet = SHIFTJIS_CHARSET;
	lstrcpyn(lf.lfFaceName, face, sizeof(lf.lfFaceName));
	
	return CreateFontIndirect(&lf);
}


int		FontCallBack(LPWMKXFDATA mkxf)
{
	
	fprintf(stderr, "\rProcessing %d/%d (%d%%)",
			mkxf->progress_count,
			mkxf->total_count,
			mkxf->progress_count * 100 / mkxf->total_count
			);
	
	if (mkxf->progress_count >= mkxf->total_count)
		fputc('\n', stderr);

	return WMKXF_NOERR;
}

int		main(int argc, char **argv)
{
	int			result = 0;
	BOOL		isprmerr = FALSE;
	LPFONTXDBCSHDR	fx;
	BOOL		isdbcs;
	int			w, h;
	char		c;
	HANDLE		hfile;
	
	if (argc < 7) {
		usage();
		return 1;
	}
	
	c = argv[2][0];
	if (c == 'S' || c == 's') isdbcs = FALSE;
	else if (c == 'D' || c == 'd') isdbcs = TRUE;
	else isprmerr = TRUE;
	
	w = atoi(argv[3]);
	h = atoi(argv[4]);
	if (w <= 0 || w > 255 || h <= 0 || h > 255) isprmerr = TRUE;
	if (isprmerr) {
		usage();
		return 1;
	}
	
	fx = makeFontxHdr(w, h, argv[5], isdbcs);
	if (!fx) {
		fprintf(stderr, "mۂłȂ悤ł\n");
		return 1;
	}
	
	hfile = CreateFile(argv[6], GENERIC_WRITE,
				0,
				NULL,
				CREATE_ALWAYS,
				FILE_ATTRIBUTE_NORMAL,
				NULL);
	if (!hfile) {
		fprintf(stderr,"t@C \"%s\" Ȃ݂ł\n", argv[6]);
	} else {
		HFONT	hfont = makeFont(argv[1], w, h, FALSE, isdbcs);
		
		if (!hfont) {
			fprintf(stderr,"tHgȂ݂ł\n");
			result = 1;
		} else {
			result = wmkxfFontToFile(fx, hfont, hfile, WMKXF_ALIGN_CENTER, FontCallBack, 0);
		}
		CloseHandle(hfile);
	}
	
	if (fx) free(fx);
	return result;
}
